from typing import Any
from typing import Callable
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from itertools import chain

import numpy as np
import tensorflow as tf

from transformers import PretrainedConfig

from nlpipes.data.data_types import InputExample


def chunk() -> Callable[[str], List[str]]:
    """  Chunk text into shorter sequences. """  
    
    def chunk_into_sentences(texts: str) -> List[str]:
        """ Chunk text with regular-expressions """
        on_regex = '(?<=[^A-Z].[.?]) +(?=[A-Z])'
        sequences = [re.split(on_regex, text) for text in texts]
        sequences = [sequence for sublist in sequences for sequence in sublist]
        return sequences
    
    def chunk_into_whatever(texts: str) -> List[str]:
        raise NotImplementedError
    
    chunk_fn = chunk_into_sentences

    return chunk_fn
    
    
def create_examples(
    texts: List[str],
    labels: Optional[List[str]],
    config: PretrainedConfig,
   ) -> List[InputExample]:
    """ Serialize each input raw data into examples. """
    
    if labels:
        label2id = config.label2id
        
        if get_labels_depth(labels)==1:
            labels = [
                label2id[label] for label in labels
            ]
        elif get_labels_depth(labels)==2:
            labels = [
                [label2id[label] for label in sub_label]
                for sub_label in labels
            ]
            
        examples = [
            InputExample(text, label=label)
            for text, label in zip(texts, labels)
        ]
        
    else:
        examples = [InputExample(text) for text in texts]
        
    return examples


def split_examples(
    examples: List[InputExample],
    train_frac: float = None,
    test_frac: float = None, 
    shufffle: bool = True,
    random_seed: int = 42,
   ) -> Tuple[InputExample]:
    """ split the input example into a training and a
    testing set. """
    
    n_examples = len(examples)
    n_train = int(np.ceil(n_examples*train_frac))
    n_test = int(np.ceil(n_examples*test_frac))
    n_total = n_train + n_test
    
    if shufffle:
        random_state = np.random.RandomState(random_seed)
        order = random_state.permutation(n_examples)
        train_idx = order[:n_train]
        test_idx = order[n_train:n_total]
    else:
        train_idx = np.arange(n_train)
        test_idx = np.arange(n_train, n_total)
    
    train_examples = [examples[idx] for idx in train_idx]
    test_examples = [examples[idx] for idx in test_idx]
    
    return train_examples, test_examples


def generate_batches(
    X: List[str], 
    Y: List[str], 
    batch_size: int,
) -> Iterable[List[Any]]:
    """ Yield a batch of examples from the input raw data """
    
    X_batch = list()
    Y_batch = list()
    
    for x, y in zip(X,Y):
        X_batch.append(x)
        Y_batch.append(y)
        if len(X_batch) < batch_size:
            continue
        yield X_batch, Y_batch    
        X_batch = list()
        Y_batch = list()


def get_labels_depth(labels: List[Any]) -> int:
    """ Get the depth level of a potentially nested list of labels """
    level = 1
    if isinstance(labels, list):
        return level + max(get_labels_depth(label) for label in labels)
        
    else:
        return 0

    
@tf.function(experimental_relax_shapes=True)
def convert_onehot_to_ids(x):
    mask = tf.dtypes.cast(x, tf.bool)
    s = tf.shape(mask)
    r = tf.reshape(
        tf.range(s[-1]),
        tf.concat([tf.ones(tf.rank(x) - 1, tf.int32), [-1]], axis=0)
    )
    r = tf.tile(r, tf.concat([s[:-1], [1]], axis=0))
    return tf.ragged.boolean_mask(r, mask)
    


# Code adapted from: https://github.com/huggingface/transformers

from typing import Tuple

import numpy as np
import tensorflow as tf

from transformers.activations_tf import ACT2FN
from transformers.activations_tf import get_tf_activation
from transformers.modeling_tf_utils import get_initializer
from transformers.modeling_tf_utils import input_processing
from transformers.modeling_tf_utils import shape_list

from nlpipes.configurations.bert_config import AdaptedBertConfig

from nlpipes.data.data_types import (
    ModelOutput,
    TFBaseModelOutput,
    TFBaseModelOutputWithPooling,
)


class TFBertEmbeddings(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)

        self.vocab_size = config.vocab_size
        self.hidden_size = config.hidden_size
        self.initializer_range = config.initializer_range
        self.position_embeddings = tf.keras.layers.Embedding(
            config.max_position_embeddings,
            config.hidden_size,
            embeddings_initializer=get_initializer(self.initializer_range),
            name="position_embeddings",
        )
        self.token_type_embeddings = tf.keras.layers.Embedding(
            config.type_vocab_size,
            config.hidden_size,
            embeddings_initializer=get_initializer(self.initializer_range),
            name="token_type_embeddings",
        )

        self.LayerNorm = tf.keras.layers.LayerNormalization(
            epsilon=config.layer_norm_eps, 
            name="LayerNorm")
        self.dropout = tf.keras.layers.Dropout(config.hidden_dropout_prob)

    def build(self, input_shape):
        """Build shared word embedding layer """
        with tf.name_scope("word_embeddings"):
            self.word_embeddings = self.add_weight(
                "weight",
                shape=[self.vocab_size, self.hidden_size],
                initializer=get_initializer(self.initializer_range),
            )

        super().build(input_shape)

    def call(self,
             input_ids=None,
             position_ids=None,
             token_type_ids=None,
             inputs_embeds=None,
             mode="embedding",
             training=False,
            ):        
        if mode == "embedding":
            return self._embedding(input_ids, position_ids, token_type_ids, 
                                   inputs_embeds, training=training)
        elif mode == "linear":
            return self._linear(input_ids)
        else:
            raise ValueError("mode {} is not valid.".format(mode))

    def _embedding(self, input_ids, position_ids, token_type_ids, 
                   inputs_embeds, training=False):
        """Applies embedding based on inputs tensor."""
        assert not (input_ids is None and inputs_embeds is None)

        if input_ids is not None:
            input_shape = shape_list(input_ids)
        else:
            input_shape = shape_list(inputs_embeds)[:-1]

        seq_length = input_shape[1]

        if position_ids is None:
            position_ids = tf.range(seq_length, dtype=tf.int32)[tf.newaxis, :]

        if token_type_ids is None:
            token_type_ids = tf.fill(input_shape, 0)

        if inputs_embeds is None:
            inputs_embeds = tf.gather(self.word_embeddings, input_ids)

        position_embeddings = tf.cast(
            self.position_embeddings(position_ids), inputs_embeds.dtype)
        token_type_embeddings = tf.cast(
            self.token_type_embeddings(token_type_ids), inputs_embeds.dtype
        )
        embeddings = inputs_embeds + position_embeddings + token_type_embeddings
        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings, training=training)

        return embeddings

    def _linear(self, inputs):
        batch_size = shape_list(inputs)[0]
        length = shape_list(inputs)[1]
        x = tf.reshape(inputs, [-1, self.hidden_size])
        logits = tf.matmul(x, self.word_embeddings, transpose_b=True)

        return tf.reshape(logits, [batch_size, length, self.vocab_size])


class TFBertSelfAttention(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)

        if config.hidden_size % config.num_attention_heads != 0:
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number of attention "
                "heads (%d)" % (config.hidden_size, config.num_attention_heads)
            )

        self.num_attention_heads = config.num_attention_heads
        assert config.hidden_size % config.num_attention_heads == 0
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size
        self.query = tf.keras.layers.Dense(
            self.all_head_size, kernel_initializer=get_initializer(config.initializer_range), name="query"
        )
        self.key = tf.keras.layers.Dense(
            self.all_head_size, kernel_initializer=get_initializer(config.initializer_range), name="key"
        )
        self.value = tf.keras.layers.Dense(
            self.all_head_size, kernel_initializer=get_initializer(config.initializer_range), name="value"
        )
        self.dropout = tf.keras.layers.Dropout(config.attention_probs_dropout_prob)

    def transpose_for_scores(self, x, batch_size):
        x = tf.reshape(x, (batch_size, -1, self.num_attention_heads, self.attention_head_size))

        return tf.transpose(x, perm=[0, 2, 1, 3])

    def call(self, hidden_states, attention_mask, head_mask, output_attentions, training=False):
        batch_size = shape_list(hidden_states)[0]
        mixed_query_layer = self.query(hidden_states)
        mixed_key_layer = self.key(hidden_states)
        mixed_value_layer = self.value(hidden_states)
        query_layer = self.transpose_for_scores(mixed_query_layer, batch_size)
        key_layer = self.transpose_for_scores(mixed_key_layer, batch_size)
        value_layer = self.transpose_for_scores(mixed_value_layer, batch_size)
        
        attention_scores = tf.matmul(
            query_layer, key_layer, transpose_b=True
        )
        dk = tf.cast(shape_list(key_layer)[-1], attention_scores.dtype)
        attention_scores = attention_scores / tf.math.sqrt(dk)
        
        if attention_mask is not None:
            attention_scores = attention_scores + attention_mask
        
        attention_probs = tf.nn.softmax(attention_scores, axis=-1)
        
        attention_probs = self.dropout(attention_probs, training=training)
        
        if head_mask is not None:
            attention_probs = attention_probs * head_mask

        context_layer = tf.matmul(attention_probs, value_layer)
        context_layer = tf.transpose(context_layer, perm=[0, 2, 1, 3])
        context_layer = tf.reshape(
            context_layer, (batch_size, -1, self.all_head_size)
        )
        outputs = (context_layer, attention_probs) if output_attentions else (context_layer,)

        return outputs


class TFAdapterLayer(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)

        self.feedforward_downproject = tf.keras.layers.Dense(
                        units=config.bottleneck_size,
                        kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=1e-3),
                        name="feedforward_downproject")
        
        self.non_linearity = ACT2FN[config.non_linearity]
    
        self.feedforward_upproject = tf.keras.layers.Dense(
                        units=config.hidden_size,
                        kernel_initializer=tf.keras.initializers.TruncatedNormal(stddev=1e-3),
                        name="feedforward_upproject")

    def call(self, hidden_states, **kwargs):

        output = self.feedforward_downproject(hidden_states)
        output = self.non_linearity(output)
        output = self.feedforward_upproject(output)
        output = output + hidden_states
        
        return output


class TFAdaptedBertSelfOutput(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)
        
        self.dense = tf.keras.layers.Dense(
            units=config.hidden_size, 
            kernel_initializer=get_initializer(config.initializer_range), 
            name="dense")
        self.dropout = tf.keras.layers.Dropout(config.hidden_dropout_prob)
        self.adapter = TFAdapterLayer(config, name="adapter")
        self.LayerNorm = tf.keras.layers.LayerNormalization(
            epsilon=config.layer_norm_eps, 
            name="LayerNorm")

    def call(self, hidden_states, input_tensor, training=False):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states, training=training)
        hidden_states = self.adapter(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        
        return hidden_states


class TFAdaptedBertAttention(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)

        self.self_attention = TFBertSelfAttention(config, name="self")
        self.dense_output = TFAdaptedBertSelfOutput(config, name="output")

    def prune_heads(self, heads):
        raise NotImplementedError

    def call(self, input_tensor, attention_mask, head_mask, output_attentions, training=False):
        self_outputs = self.self_attention(
            input_tensor, attention_mask, head_mask, output_attentions, training=training
        )
        attention_output = self.dense_output(self_outputs[0], input_tensor, training=training)
        outputs = (attention_output,) + self_outputs[1:]

        return outputs


class TFBertIntermediate(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)

        self.dense = tf.keras.layers.Dense(
            config.intermediate_size, kernel_initializer=get_initializer(config.initializer_range), name="dense"
        )

        if isinstance(config.hidden_act, str):
            self.intermediate_act_fn = get_tf_activation(config.hidden_act)
        else:
            self.intermediate_act_fn = config.hidden_act

    def call(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.intermediate_act_fn(hidden_states)

        return hidden_states


class TFAdaptedBertOutput(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)
        
        self.dense = tf.keras.layers.Dense(
            config.hidden_size, 
            kernel_initializer=get_initializer(config.initializer_range),
            name="dense")
        self.dropout = tf.keras.layers.Dropout(config.hidden_dropout_prob)
        self.adapter = TFAdapterLayer(config, name="adapter")
        self.LayerNorm = tf.keras.layers.LayerNormalization(
            epsilon=config.layer_norm_eps,
            name="LayerNorm")
        
    def call(self, hidden_states, input_tensor, training=False):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states, training=training)
        hidden_states = self.adapter(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        
        return hidden_states


class TFAdaptedBertLayer(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)

        self.attention = TFAdaptedBertAttention(config, name="attention")
        self.intermediate = TFBertIntermediate(config, name="intermediate")
        self.bert_output = TFAdaptedBertOutput(config, name="output")

    def call(self, hidden_states, attention_mask, head_mask, output_attentions, training=False):
        attention_outputs = self.attention(
            hidden_states, attention_mask, head_mask, output_attentions, training=training
        )
        attention_output = attention_outputs[0]
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.bert_output(
            intermediate_output, attention_output, training=training,
        )
        outputs = (layer_output,) + attention_outputs[1:]

        return outputs


class TFAdaptedBertEncoder(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)

        self.layer = [TFAdaptedBertLayer(config, name="layer_._{}".format(i)) for \
                      i in range(config.num_hidden_layers)]
    
    def call(
        self,
        hidden_states,
        attention_mask,
        head_mask,
        output_attentions,
        output_hidden_states,
        return_dict,
        training=False,
    ):
        all_hidden_states = () if output_hidden_states else None
        all_attentions = () if output_attentions else None

        for i, layer_module in enumerate(self.layer):
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            layer_outputs = layer_module(
                hidden_states, attention_mask, head_mask[i], output_attentions, training=training
            )
            hidden_states = layer_outputs[0]

            if output_attentions:
                all_attentions = all_attentions + (layer_outputs[1],)

        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if not return_dict:
            return tuple(v for v in [hidden_states, all_hidden_states, all_attentions] if v is not None)
        
        return TFBaseModelOutput(
            last_hidden_state=hidden_states, hidden_states=all_hidden_states, attentions=all_attentions
        )


class TFBertPooler(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)

        self.dense = tf.keras.layers.Dense(
            config.hidden_size,
            kernel_initializer=get_initializer(config.initializer_range),
            activation="tanh",
            name="dense",
        )

    def call(self, hidden_states):
        first_token_tensor = hidden_states[:, 0]
        pooled_output = self.dense(first_token_tensor)

        return pooled_output


class TFAdaptedBertMainLayer(tf.keras.layers.Layer):
    config_class = AdaptedBertConfig

    def __init__(self, config,
                 add_pooling_layer=True, 
                 **kwargs):
        super().__init__(**kwargs)

        self.config = config
        self.num_hidden_layers = config.num_hidden_layers
        self.initializer_range = config.initializer_range
        self.output_attentions = config.output_attentions
        self.output_hidden_states = config.output_hidden_states
        self.return_dict = config.use_return_dict
        self.embeddings = TFBertEmbeddings(config, name="embeddings")
        self.encoder = TFAdaptedBertEncoder(config, name="encoder")
        self.pooler = TFBertPooler(config, name="pooler") if add_pooling_layer else None
        self.set_trainable_variables(config)
    
    def get_input_embeddings(self):
        return self.embeddings

    def set_input_embeddings(self, value):
        self.embeddings.word_embeddings = value
        self.embeddings.vocab_size = value.shape[0]

    def set_trainable_variables(self, config):
        """ When adapter layers are added, only the adapter layers and the
        normalization layers are updated during the training process """       
        self.embeddings.trainable = False
        self.pooler.trainable = False
        for index in range(config.num_hidden_layers):
                self.encoder.layer[index].attention.self_attention.trainable = False
                self.encoder.layer[index].attention.dense_output.dense.trainable = False
                self.encoder.layer[index].attention.dense_output.adapter.trainable = True
                self.encoder.layer[index].attention.dense_output.LayerNorm.trainable = True
                self.encoder.layer[index].intermediate.trainable = False
                self.encoder.layer[index].bert_output.dense.trainable = False
                self.encoder.layer[index].bert_output.adapter.trainable = True 
                self.encoder.layer[index].bert_output.LayerNorm.trainable = True

    def call(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        training=False,
        **kwargs,
    ):
        inputs = input_processing(
            func=self.call,
            config=self.config,
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            training=training,
            kwargs_call=kwargs,
        )

        if inputs["input_ids"] is not None and inputs["inputs_embeds"] is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif inputs["input_ids"] is not None:
            input_shape = shape_list(inputs["input_ids"])
        elif inputs["inputs_embeds"] is not None:
            input_shape = shape_list(inputs["inputs_embeds"])[:-1]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        if inputs["attention_mask"] is None:
            inputs["attention_mask"] = tf.fill(input_shape, 1)

        if inputs["token_type_ids"] is None:
            inputs["token_type_ids"] = tf.fill(input_shape, 0)
            
        embedding_output = self.embeddings(
            inputs["input_ids"],
            inputs["position_ids"],
            inputs["token_type_ids"],
            inputs["inputs_embeds"],
            training=inputs["training"],
        )
        
        extended_attention_mask = inputs["attention_mask"][:, tf.newaxis, tf.newaxis, :]
        extended_attention_mask = tf.cast(extended_attention_mask, embedding_output.dtype)
        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0

        if inputs["head_mask"] is not None:
            raise NotImplementedError
        else:
            inputs["head_mask"] = [None] * self.num_hidden_layers

        encoder_outputs = self.encoder(
            embedding_output,
            extended_attention_mask,
            inputs["head_mask"],
            inputs["output_attentions"],
            inputs["output_hidden_states"],
            inputs["return_dict"],
            training=inputs["training"],
        )

        sequence_output = encoder_outputs[0]
        pooled_output = self.pooler(sequence_output) if self.pooler is not None else None

        if not inputs["return_dict"]:
            return (
                sequence_output,
                pooled_output,
            ) + encoder_outputs[1:]

        return TFBaseModelOutputWithPooling(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
        )



from typing import Callable
from typing import List

from dataclasses import dataclass

import copy

import numpy as np
import tensorflow as tf

from transformers import (
    PretrainedConfig,
    PreTrainedTokenizer,
    AutoTokenizer,
)

from nlpipes.data.data_types import InputExample
from nlpipes.data.data_types import InputFeatures


""" The data processor classes.

The data processor encapsulates all the transformations
needed to encode input examples into features to be used
as input batch by the model. """


@dataclass
class LanguageModelingDataProcessor:
    """ 
    Data processor for Masked Language Modeling:
    80% MASK, 10% random, 10% original.
    Inputs are dynamically padded to the maximum 
    length of a batch if they are not all of the 
    same length.
    
    Args
    ----------

    tokenizer(PreTrainedTokenizer):
        The tokenizer used for encoding the data.

    config(PreTrainedCOnfig):
        The configuration of the model.
    """
    
    name: str = 'LanguageModelingProcessor'
    tokenizer: PreTrainedTokenizer = None
    config: PretrainedConfig = None
    mlm_proba: float = None
    
    def __call__(self, 
                 examples: List[InputExample],
                ) -> InputFeatures:
        
        texts = [example.text for example in examples]
        
        inputs = self.tokenizer(
            texts,
            add_special_tokens=True,
            padding=True,
            truncation=True,
            max_length=512,
            is_split_into_words=False,
            return_attention_mask=True,
            return_token_type_ids=True,
            return_tensors='tf',
        )
        
        input_ids = tf.identity(inputs.input_ids)
        
        mask = self.create_mask(
            inputs=inputs.input_ids,
            mask_frac=self.mlm_proba,
            excluded=self.tokenizer.all_special_ids,
        )
        
        inputs['input_ids'] = self.apply_mask(
            inputs=inputs.input_ids,
            mask=mask,
            mask_token_id=self.tokenizer.mask_token_id,
        )
        
        labels = self.apply_mask(
            inputs=input_ids,
            mask=~mask,
            mask_token_id=-100,
        )
        
        inputs['label'] = tf.one_hot(
            [label for label in labels],
            depth=len(self.tokenizer)
        )
        
        input_features = InputFeatures(
            input_ids=inputs['input_ids'],
            attention_mask=inputs['attention_mask'],
            token_type_ids=inputs['token_type_ids'],
            label=inputs['label'],
        )
        
        return input_features
    
    def create_mask(
        self,
        inputs: tf.Tensor,
        mask_frac: float,
        excluded: tf.Tensor,  
    ) -> tf.Tensor:
        """ Create the mask used for the masked 
        language modeling """
        
        random = tf.random.uniform(inputs.shape)
        
        mask = tf.math.logical_and(
            random < mask_frac, 
            ~np.isin(inputs, excluded),
        )
        
        return mask
    
    def apply_mask(
        self,
        inputs: tf.Tensor,
        mask: tf.Tensor,
        mask_token_id: int,
    ) -> tf.Tensor:
        """ Apply the mask to the given inputs """
        
        masked_inputs = tf.where(
            condition=mask,
            x=mask_token_id,
            y=inputs,
        )
        
        return masked_inputs


@dataclass
class SingleLabelDataProcessor:
    """
    Data Processor for single-label sequence classification.
    Inputs are dynamically padded to the maximum 
    length of a batch if they are not all of the 
    same length.
    
    Args
    ----------

    tokenizer (transformers.PreTrainedTokenizer):
        The tokenizer used for encoding the data.

    config (transformers.PreTrainedCOnfig):
        The configuration of the model.
    """
    
    name: str = 'SingleLabelClassificationProcessor'
    tokenizer: PreTrainedTokenizer = None
    config: PretrainedConfig = None
    seq_max_len: int = 512
    
    def __call__(self, 
                 examples: List[InputExample],
                ) -> InputFeatures:
        
        texts = [example.text for example in examples]
        
        inputs = self.tokenizer(
            texts,
            add_special_tokens=True,
            padding=True,
            truncation=True,
            max_length=self.seq_max_len,
            is_split_into_words=False,
            return_attention_mask=True,
            return_token_type_ids=True,
            return_tensors='tf',
        )
        
        inputs['label'] = tf.one_hot(
            [example.label for example in examples],
            depth=self.config.num_labels
        )
        
        input_features = InputFeatures(
            input_ids=inputs['input_ids'],
            attention_mask=inputs['attention_mask'],
            token_type_ids=inputs['token_type_ids'],
            label=inputs['label'],
        )
        
        return input_features


@dataclass(frozen=True)
class MultiLabelDataProcessor:
    """
    Data Processor for multi-label sequence classification.
    Inputs are dynamically padded to the maximum length of a
    batch if they are not all of the same length.
    
    Args
    ----------

    tokenizer (transformers.PreTrainedTokenizer):
        The tokenizer used for encoding the data.

    config (transformers.PreTrainedCOnfig):
        The configuration of the model.
    """
    
    name: str = 'MultiLabelClassificationProcessor'
    tokenizer: PreTrainedTokenizer = None
    config: PretrainedConfig = None
    seq_max_len: int = 512
    
    def __call__(self, 
                 examples: List[InputExample],
                ) -> InputFeatures:
        
        texts = [example.text for example in examples]
        
        inputs = self.tokenizer(
            texts,
            add_special_tokens=True,
            padding=True,
            truncation=True,
            max_length=self.seq_max_len,
            is_split_into_words=False,
            return_attention_mask=True,
            return_token_type_ids=True,
            return_tensors='tf',
        )
        
        labels = [example.label for example in examples]
        ragged_labels = tf.ragged.constant(labels)	
        onehot_labels = tf.one_hot(
            ragged_labels,
            depth=self.config.num_labels,
        )
        
        inputs['label'] = tf.reduce_max(onehot_labels, axis=1)
        
        input_features = InputFeatures(
            input_ids=inputs['input_ids'],
            attention_mask=inputs['attention_mask'],
            token_type_ids=inputs['token_type_ids'],
            label=inputs['label'],
        )
        
        return input_features
    


import os
import logging
import shutil

from typing import List
from typing import Optional

from dataclasses import dataclass

from pathlib import Path

import tensorflow as tf
import transformers

from transformers import (
    PretrainedConfig,
    PreTrainedTokenizerFast,
    TFPreTrainedModel,
    TFAutoModelForMaskedLM,
    AutoConfig,
    AutoTokenizer,
)

from tokenizers.pre_tokenizers import Whitespace

from nlpipes.trainers.trainers import Trainer
from nlpipes.losses.losses import softmax_cross_entropy

from nlpipes.data.data_loaders import DataLoader
from nlpipes.data.data_selectors import DataSelector
from nlpipes.data.data_augmentors import VocabAugmentor
from nlpipes.data.data_processors import SingleLabelDataProcessor
from nlpipes.data.data_cleaners import clean

from nlpipes.data.data_utils import (
     create_examples,
     split_examples,
     generate_batches,
     chunk,
)

from nlpipes.data.data_types import (
     Corpus,
     Document,
     Token,
     InputExample,
     InputExamples,
     InputTokens,
     InputFeatures,
     Predictions,
     Outcomes,
)

from nlpipes.callbacks.callbacks import (
     Callback,
     TrainingStep,
     History,
     ModelCheckpoint,
     CSVLogger,
     ProgbarLogger,
     EarlyStopping,
     TimedStopping,
)

transformers.logging.set_verbosity_error()

logger = logging.getLogger('__name__')

# uncomment row below for debugging rust tokenizer
# os.environ["RUST_BACKTRACE"]="full" 


@dataclass
class TFPipelineForSingleLabelClassification():
    """ The pipeline for single-label sequence 
    classification. There are almost no abstractions. 
    Just three standard classes are required to use 
    pipeline: configuration, model and tokenizer.
    
    Args
    ----------
    model(TFPreTrainedModel):
       The model used for the classification task`.
       It is simply a bert layer with a classification
       head on top.
    
    tokenizer(PreTrainedTokenizerFast):
       The rust tokenizer used to convert the input
       text into encoded vectors.
    
    config(PretrainedConfig): 
       The model configuration file.
    """
    
    model: TFPreTrainedModel
    tokenizer: PreTrainedTokenizerFast
    config: PretrainedConfig
    
    def train(self, 
              X: List[str], 
              Y: List[str],
              train_frac: float = 0.75,
              test_frac: float = 0.25,
              num_epochs: int = 3,
              batch_size: int = 16,
              seq_max_len: int = 512,
              learning_rate: float = 2e-5,
              beta_1: float = 0.9,
              beta_2: float = 0.999,
              patience: int = 2,
              min_delta: float = 0.01,
              time_limit: int = 86400,
              random_seed: int = 42,
              keep_checkpoints: bool = False,
              checkpoints_dir: str = './checkpoints',
              logs_dir: str = './logs',
             ) -> List[History]:
        """ Train a model for single-label classification.
        Most of the training process is implemented through 
        callbacks functions. Even the actual training step 
        (calculating the gradient and updating the weights 
        using an optimizer) is implemented as callbacks and
        is not part of the Trainer itself. This provide a 
        modular architecture that allows new ideas to be 
        implemented without having to change and increase
        the complexity of the trainer. """
        
        examples = create_examples(
            texts=X, 
            labels=Y,
            config=self.config,
        )
        
        train_examples, test_examples = split_examples(
            examples,
            train_frac=train_frac,
            test_frac=test_frac,
            shufffle=True,
            random_seed=random_seed,
        )          
        
        data_processor = SingleLabelDataProcessor(
            name='DataProcessing',
            tokenizer=self.tokenizer,
            config=self.config,
            seq_max_len=seq_max_len,
        )
        
        train_stream = DataLoader(
            name='TrainDataLoader',
            examples=train_examples,
            batch_size=batch_size,
            data_processor=data_processor,
            seed=random_seed,
        )
        
        test_stream = DataLoader(
            name='TestDataLoader',
            examples=test_examples,
            batch_size=batch_size,
            data_processor=data_processor,
            seed=random_seed,
        )
        
        training_step = TrainingStep(
            name='TrainingStep',
            model=self.model,
            loss_function=softmax_cross_entropy,
            optimizer=tf.optimizers.experimental.AdamW(
                learning_rate=learning_rate,
                beta_1=beta_1,
                beta_2=beta_2,
            ),
        )
        
        history = History(
            name='History', 
            training_step=training_step,
            loss_metric=tf.metrics.Mean,
            acc_metric=tf.metrics.CategoricalAccuracy(),
        )
        
        model_checkpoint = ModelCheckpoint(
            name='ModelCheckpoint',
            model=self.model, 
            history=history,
            checkpoints_dir=checkpoints_dir,
        )
        
        early_stopping = EarlyStopping(
            name='EarlyStopping', 
            history=history,
            direction='minimize',
            min_delta=min_delta,
            patience=patience,
        )
        
        timed_stopping = TimedStopping(
            name='TimedStopping',
            time_limit=time_limit,
        )
        
        csv_logger = CSVLogger(
            name='CSVLogging',
            logs_dir=logs_dir,
        )
        
        progbar_logger = ProgbarLogger(
            name='ProgbarLogger',
            batch_size=batch_size,
            num_epochs=num_epochs,
            history=history,
            num_samples=sum([
                train_stream.num_examples(),
                test_stream.num_examples(),
            ]))
        
        callbacks = [
            training_step,
            history,
            model_checkpoint,
            csv_logger,
            progbar_logger,
            early_stopping,
            timed_stopping,
        ]
        
        trainer = Trainer(
            train_stream=train_stream,
            test_stream=test_stream,
            callbacks=callbacks,
            num_epochs=num_epochs,
        )
        
        trainer.train()
        
        if keep_checkpoints == False:
            shutil.rmtree(checkpoints_dir)
            
        return history
    
    def evaluate(self,
                 X: List[str],
                 Y: List[str],
                 metric: tf.metrics.Metric,
                 batch_size: int = 32,
                ) -> tf.metrics.Metric:
        """ Evaluate the model performance on new data
        according to the user-defined evaluation metric """
        
        for texts, labels in generate_batches(X, Y, batch_size):
            examples = self.preprocess(texts, labels)
            tokens = self.tokenize(examples)
            features = self.encode(tokens)
            predictions = self.classify(features)
            
            y_true = examples.labels
            y_pred = predictions.labels
            
            metric.update_state(y_true, y_pred)
        
        return metric.result()
    
    def predict(self, 
                texts: Corpus, 
                chunk_text: bool = False,
               ) -> Outcomes:
        """ Get predictions on new unlabeled data """
         
        examples = self.preprocess(texts)
        tokens = self.tokenize(examples)
        features = self.encode(tokens)
        predictions = self.classify(features)
        completions = self.postprocess(predictions)
        
        return completions
    
    def preprocess(self,
                   texts: Corpus,
                   labels: Optional[str] = None,
                   chunk_text: Optional[bool] = False,
                  ) -> InputExamples:
        """ Serialize each raw text into examples. 
        Raw texts can eventually be chunk into spans """
        
        if labels:
            label2id = self.config.label2id
            labels_id = [label2id[label] for label in labels]
            examples = InputExamples(texts, labels=labels_id)
        else:
            texts = chunk(text) if chunk_text else texts
            examples = InputExamples(texts)
        
        return examples

    def tokenize(self, examples: InputExamples) -> InputTokens:
        """ Convert examples into tokens. Tokens encapsulate both 
        words, subwords and enriched subwords with the special
        [CLS] and [SEP] tokens. """
        
        texts = examples.texts
        
        words = [
            Whitespace().pre_tokenize_str(text) 
            for text in texts
        ]

        subwords = [
            encoding.tokens for encoding 
            in self.tokenizer.backend_tokenizer.encode_batch(
               texts, add_special_tokens=False)
        ]

        subwords_plus = [
            encoding.tokens for encoding 
            in self.tokenizer.backend_tokenizer.encode_batch(
               texts, add_special_tokens=True)
        ]

        tokens = InputTokens(
            words=words,
            subwords=subwords,
            subwords_plus=subwords_plus,
        )

        return tokens

    def encode(self, tokens: InputTokens) -> InputFeatures: 
        """ Encode human-readable subwords into encoded
        features that can be use as input by the model. """

        inputs = self.tokenizer(
           tokens.subwords,
           add_special_tokens=False,
           padding=True,
           truncation=True,
           max_length=160,
           is_split_into_words = True,
           return_attention_mask=True,
           return_token_type_ids=True,
           return_tensors='tf',
        )

        features = InputFeatures(
           input_ids=inputs['input_ids'],
           attention_mask=inputs['attention_mask'],
           token_type_ids=inputs['token_type_ids'],
        )

        return features

    def classify(self, features: InputFeatures) -> Predictions:
        """ Predict the label index matching the maximum 
        prediction score and compute the loss value associated
        with each prediction. """

        outputs = self.model.call(
            input_ids=features.input_ids,
            attention_mask=features.attention_mask,
            token_type_ids=features.token_type_ids,
        )

        logits = outputs.logits
        scores = tf.nn.softmax(logits, axis=1)
        labels = tf.argmax(logits, axis=-1)
        labels_onehot = tf.one_hot(labels, self.config.num_labels)
        losses = softmax_cross_entropy(labels_onehot, logits)
        
        predictions = Predictions(
            labels=labels,
            scores=scores,
            losses=losses,
        )
        
        return predictions

    def postprocess(self, predictions: Predictions) -> Outcomes:
        """ Post-process the predictions. We convert tensors 
        into more convenient python lists and also convert the 
        predicted labels ids into their original name. """
          
        labels = predictions.labels.numpy()
        scores = predictions.scores.numpy()
        losses = predictions.losses.numpy()

        labels = labels.tolist()
        scores = scores.tolist()
        losses = losses.tolist()
        
        id2label = self.config.id2label
        labels = [id2label[label] for label in labels]
        
        outcomes = Outcomes(
            labels=labels,
            scores=scores,
            losses=losses,
        )

        return outcomes
    
    def save(self, save_dir: str):
        """ Save the trained model on disk """
        
        os.makedirs(save_dir, exist_ok=True)
        self.model.save_pretrained(save_dir)
        self.tokenizer.save_pretrained(save_dir)
        self.config.save_pretrained(save_dir)



from typing import Union
from typing import Optional

from typeguard import typechecked

import numpy as np
import tensorflow as tf


class ConfusionMatrix(tf.metrics.Metric):
    """ Compute the confusion matrix for single-label
    classifier. 
    
    Args:
    ---------
    num_classes: (int): the number of predictions labels
    name: (Optional) name of the metric instance.
    dtype: (Optional) data type of the metric result.
    """
    
    @typechecked
    def __init__(
        self, 
        num_classes: int,
        name: str = "multiclass_confusion_matrix",
        dtype = tf.dtypes.int32,
    ):
        super().__init__(name=name, dtype=dtype)
        self.num_classes = num_classes
        self.confusion_matrix = self.add_weight(
            'confusion_matrix',
            shape=[num_classes, num_classes],
            initializer='zeros',
            dtype=self.dtype,
        )

    def update_state(self, y_true, y_pred):
        """ Accumulates statistics for the metric. """
        batch = tf.math.confusion_matrix(
            y_true, y_pred,
            num_classes=self.num_classes,
            dtype=self.dtype,
        )
        self.confusion_matrix.assign_add(batch)

    def result(self):
        """ Return the metric result"""
        return self.confusion_matrix


class MultiLabelConfusionMatrix(tf.metrics.Metric):
    """ Computes the confusion matrix for multi-labels 
    classifier. Every class has a dedicated 
    matrix of shape `(2, 2)` that contains:
    - true negatives in cell `(0,0)`
    - false positives in cell `(0,1)`
    - false negatives in cell `(1,0)`
    - true positives in cell `(1,1)`
    
    Args:
    ---------
    num_classes: (int): the number of predictions labels
    name: (Optional) name of the metric instance.
    dtype: (Optional) data type of the metric result.
    """

    @typechecked
    def __init__(
        self,
        num_classes: int,
        name: str = "multilabel_confusion_matrix",
        dtype = tf.dtypes.int32,
        **kwargs,
    ):
        super().__init__(name=name, dtype=dtype)
        self.num_classes = num_classes
        self.true_positives = self.add_weight(
            "true_positives",
            shape=[self.num_classes],
            initializer="zeros",
            dtype=self.dtype,
        )
        self.false_positives = self.add_weight(
            "false_positives",
            shape=[self.num_classes],
            initializer="zeros",
            dtype=self.dtype,
        )
        self.false_negatives = self.add_weight(
            "false_negatives",
            shape=[self.num_classes],
            initializer="zeros",
            dtype=self.dtype,
        )
        self.true_negatives = self.add_weight(
            "true_negatives",
            shape=[self.num_classes],
            initializer="zeros",
            dtype=self.dtype,
        )

    def update_state(self, y_true, y_pred):
        """ Accumulates statistics for the metric. """
        y_true = tf.cast(y_true, self.dtype)
        y_pred = tf.cast(y_pred, self.dtype)
        
        pred_sum = tf.math.count_nonzero(y_pred, 0)
        true_sum = tf.math.count_nonzero(y_true, 0)
        
        y_true_negative = tf.math.not_equal(y_true, 1)
        y_pred_negative = tf.math.not_equal(y_pred, 1)
        
        true_positive = tf.math.count_nonzero(y_true * y_pred, 0)
        false_positive = pred_sum - true_positive
        false_negative = true_sum - true_positive
        true_negative = tf.math.count_nonzero(
            tf.math.logical_and(y_true_negative, y_pred_negative), 
            axis=0,
        )

        self.true_positives.assign_add(tf.cast(true_positive, self.dtype))
        self.false_positives.assign_add(tf.cast(false_positive, self.dtype))
        self.false_negatives.assign_add(tf.cast(false_negative, self.dtype))
        self.true_negatives.assign_add(tf.cast(true_negative, self.dtype))

    def result(self):
        """ Returns the metric value. """
        flat_confusion_matrix = tf.convert_to_tensor(
            [
                self.true_negatives,
                self.false_positives,
                self.false_negatives,
                self.true_positives,
            ]
        )
        
        confusion_matrix = tf.reshape(
            tf.transpose(flat_confusion_matrix), [-1, 2, 2]
        )

        return confusion_matrix
    
    def reset_state(self):
        """ Resets all of the metric state variables. """
        reset_value = np.zeros(self.num_classes, dtype=np.int32)
        tf.keras.backend.batch_set_value(
            [(v, reset_value) for v in self.variables]
        )
    
    
def hamming_loss_with_logits(
    y_true: tf.Tensor,
    y_pred: tf.Tensor,
    threshold: Union[tf.Tensor, None],
) -> tf.Tensor:
    """  Computes the Hamming loss """

    #convert logits to `probability` scores
    y_pred = tf.sigmoid(tf.cast(y_pred, tf.float32))
    
    if threshold is None:
        threshold = tf.reduce_max(y_pred, axis=-1, keepdims=True)
        y_pred = tf.logical_and(
            y_pred >= threshold, 
            tf.abs(y_pred) > 1e-12,
        )
    else:
        y_pred = y_pred > threshold

    y_true = tf.cast(y_true, tf.int32)
    y_pred = tf.cast(y_pred, tf.int32)
    
    mismatches = tf.cast(
        tf.math.count_nonzero(y_true - y_pred, axis=-1),
        tf.float32,
    )
     
    return 1 - mismatches / y_true.get_shape()[-1]
    
    
class HammingLoss(tf.metrics.MeanMetricWrapper):
    """ Compute the average Hamming loss between y_true and y_pred.
    The Hamming loss is the fraction of wrongly predicted labels to
    the total number of labels.
    
    Args:
    ----------
        y_true: actual target value.
        y_pred: predicted target value.
        threshold: Elements of `y_pred` greater than threshold
            are converted to be 1, and the rest 0. If threshold
            is None, the argmax is converted to 1, and the rest 0.
    Returns:
        hamming loss: float.
    """

    @typechecked
    def __init__(
        self,
        name: str = "hamming_loss",
        threshold: Union[tf.Tensor, float] = None,
        dtype = tf.float32,
        **kwargs,
    ):
        super().__init__(
            fn=hamming_loss_with_logits, 
            name=name, 
            dtype=dtype,
            threshold=threshold,
        )
        


# Code lightly adapted from https://github.com/georgian-io/Transformers-Domain-Adaptation

from collections import Counter

from typing import List
from typing import Optional
from typing import Union
from typing import Sequence
from typing import NewType

from tqdm import tqdm

import numpy as np
import pandas as pd
from scipy import sparse

from sklearn.preprocessing import RobustScaler

from transformers import PreTrainedTokenizer

from nlpipes.data.data_types import (
     Corpus,
     Token,
)

from nlpipes.metrics import (
    DIVERSITY_FEATURES,
    SIMILARITY_FEATURES,
    diversity_metrics_factory,
    similarity_metrics_factory,
)

np.seterr(invalid='ignore')


class DataSelector():
    """ """
    def __init__(
        self,
        name: str,
        target_corpus_size: Union[int, float],
        tokenizer: PreTrainedTokenizer,
        similarity_metrics: Optional[List[str]] = None,
        diversity_metrics: Optional[List[str]] = None,
    ):
        self.target_corpus_size = target_corpus_size
        self.tokenizer = tokenizer
        self.similarity_metrics = similarity_metrics
        self.diversity_metrics = diversity_metrics

    def __call__(self, corpus: Corpus) -> Corpus:
        self.fit(corpus)
        sub_corpus = self.select(corpus)
        return sub_corpus
    
    def fit(self, corpus: Corpus):
        """ Get tokens distribution from the corpus """
        self.tokens_distribution = self.get_tokens_distribution(" ".join(corpus))
        return self
    
    def select(self, corpus: Corpus) -> Corpus:
        """ Select the most relevant texts from the corpus
        according to the provided selection metrics """
        
        scores = self.compute_metrics(corpus)
        composite_scores = scores["composite"].sort_values(ascending=False)
        n_selected = (self.target_corpus_size 
                      if isinstance(self.target_corpus_size, int)
                      else int(self.target_corpus_size * len(corpus))
                     )
        selection_index = composite_scores.index[:n_selected]
        
        sub_corpus = pd.Series(corpus)[selection_index]
        
        return sub_corpus.tolist()

    def compute_metrics(self, corpus: Corpus) -> pd.DataFrame:
        """ """
        scores = pd.concat([self.compute_similarities(corpus),
                            self.compute_diversities(corpus),
                           ],
                           axis=1,
                          )
        
        # Ensure metrics are normalized
        scores = pd.DataFrame(
            RobustScaler().fit_transform(scores), columns=scores.columns
        )
        
        scores["composite"] = scores.sum(axis=1)
        return scores
    
    def compute_similarities(self, corpus: Corpus) -> pd.DataFrame:
        """ """ 
        
        similarities = pd.DataFrame()
        
        if (self.similarity_metrics is None):  
            return similarities

        tokens_distribution = self.get_tokens_distribution_batch(corpus)

        progress_bar = tqdm(
            self.similarity_metrics,
            desc="computing similarity",
            unit="metric",
            dynamic_ncols=True,
        )
        
        for metric in progress_bar:
            similarity_func = similarity_metrics_factory(metric)
            similarities[metric] = similarity_func(
                tokens_distribution, 
                self.tokens_distribution.reshape(1, -1),
            )
        
        return similarities
    
    def compute_diversities(self, corpus: Corpus) -> pd.DataFrame:
        """ """ 
        diversities = pd.DataFrame()
        
        if self.diversity_metrics is None:
            return diversities
        
        tokenized_corpus = self.tokenizer(corpus, add_special_tokens=False)
        tokens = [
            self.tokenizer.convert_ids_to_tokens(token_ids) 
            for token_ids in tokenized_corpus.input_ids
        ]
        
        progress_bar = tqdm(
            self.diversity_metrics,
            desc="computing diversity",
            unit="metric",
            dynamic_ncols=True,
        )
        
        for metric in progress_bar:
            diversity_function = diversity_metrics_factory(
                metric,
                tokens_distribution=self.tokens_distribution,
                vocab2id=self.tokenizer.vocab,
            )
            diversities[metric] = pd.Series(
                diversity_function(token) 
                for token in tokens
            )

        return diversities
    
    def get_tokens_distribution(self, text: str) -> np.ndarray:
        """ Get the vocabulary distribution of a corpus. The 
        function computes the token frequency for each corpus
        token detected in the tokenizer vocabulary. """
        
        vocabulary = self.tokenizer.vocab
        
        tokens = self.tokenizer.tokenize(text)
        tokens_count = Counter(tokens)
        
        tokens_distribution = np.zeros(len(vocabulary))
        for token, count in tokens_count.items():
            token_idx = vocabulary[token]
            tokens_distribution[token_idx] = count
        tokens_distribution /= tokens_distribution.sum()
        
        return tokens_distribution

    def get_tokens_distribution_batch(self, corpus: Corpus) -> np.ndarray:
        """ """
        
        counters = [
            Counter(encoding.ids)
            for encoding in self.tokenizer.backend_tokenizer.encode_batch(
                corpus, add_special_tokens=False
            )
        ]
        
        rows = np.array(
            [
                val for i, 
                counter in enumerate(counters) for val in [i] * len(counter)
            ]
        )
        cols = np.array(
            [
                key for counter in counters 
                for key in counter.keys()
            ]
        )
        data = np.array(
            [
                value for counter in counters 
                for value in counter.values()
            ]
        )
        
        term_counts = sparse.csr_matrix(
            (data, (rows, cols)),
            shape=(len(counters), len(self.tokenizer)),
            dtype=np.uint16 if len(self.tokenizer) < 2**16 else np.uint32,
        )

        term_dist = term_counts / term_counts.sum(axis=1)
        return np.array(term_dist)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nlpipes',
 'nlpipes.callbacks',
 'nlpipes.configurations',
 'nlpipes.data',
 'nlpipes.layers',
 'nlpipes.losses',
 'nlpipes.metrics',
 'nlpipes.models',
 'nlpipes.optimization',
 'nlpipes.pipelines',
 'nlpipes.trainers']

package_data = \
{'': ['*']}

install_requires = \
['ftfy>=6.1.1,<7.0.0',
 'numpy>=1.21.0,<2.0.0',
 'pandas>=1.3.5,<2.0.0',
 'scipy>=1.7.3,<2.0.0',
 'tensorflow>=2.11.0,<3.0.0',
 'tokenizers>=0.13.0,<0.14.0',
 'tqdm>=4.64.0,<5.0.0',
 'transformers>=4.24.0,<5.0.0']

setup_kwargs = {
    'name': 'nlpipes',
    'version': '0.1.12',
    'description': 'Text Classification with Transformers',
    'long_description': '<!-- PROJECT NAME -->\n<div align="center">\n   <img src="https://raw.githubusercontent.com/uyanik/Images/b90c13ff8e580b2ecb2468889ed5ab1d52633d2e/logo.png" alt="nlpipes_logo" title="nlpipes logo">\n  <h2>Text Classification with Transformers</h2>\n</div>\n\n<div align="center">\n    <a href="https://opensource.org/licenses/Apache-2.0">\n       <img alt="Licence" src="https://img.shields.io/badge/License-Apache_2.0-blue.svg">\n    </a>\n     <a href="https://pypi.org/project/nlpipes/">\n       <img alt="PyPi Version" src="https://img.shields.io/pypi/pyversions/nlpipes">\n    </a> \n    <a href="https://pypi.org/project/nlpipes/">\n        <img alt="PyPi Package Version" src="https://img.shields.io/pypi/v/nlpipes">\n    </a>\n    <!--\n    <a href="https://pepy.tech/project/nlpipes/">\n        <img alt="PyPi Downloads" src="https://static.pepy.tech/badge/nlpipes/month">\n    </a>\n    -->\n</div>\n\n<div align="center">\n    <a href=""><strong>Documentation</strong></a>\n    • <a href=""><strong>References</strong></a>\n</div>\n\n\n<div align="center">\n  <img src="https://ik.imagekit.io/m0ci8dgk4/nlpipes_screenshot_w_309qBjx.png?ik-sdk-version=javascript-1.4.3&updatedAt=1678033602383" alt="nlpipes_screenshot" title="nlpipes screenshot">\n</div>\n\n\n## Overview\n`NLPipes` provides an easy way to use Transformers-based models for training, evaluation and inference on a diversity of text classification tasks, including:\n\n* **Single-label classification**: Assign one label to each text. A typical use case is sentiment detection where one want to detect the overall sentiment polarity (e.g., positive, neutral, negative) in a review.\n* **Multi-labels classification** [Not yet implemented]: Assign one or more label to each text from a list of possible labels. A typical use case is aspect categories detection where one want to detect the multiple aspects mentionned in a review (e.g., product_quality, delivery_time, price, ...).\n* **Aspect-based classification** [Not yet implemented]: Assign one label from a list of possible labels for each of a list of aspects. A typical use case is aspect based sentiment analysis where one want to detect each aspect categories mentionned in a review along his assocated sentiment polarity (e.g., product_quality: neutral, delivery_time: negative, price: positive, ...).\n\n`NLPipes` expose a simple `Model` API that offers a common abstraction to run several text classification tasks. The `Model` encapsulate most of the complex code from the library and save having to deal with the complexity of transformers based algorithms.\n\n#### Built with\n`NLPipes` is built with TensorFlow and HuggingFace Transformers:\n* [TensorFlow](https://www.tensorflow.org/): An end-to-end open source deep learning framework\n* [Transformers](https://huggingface.co/transformers/): An general-purpose open-sources library for transformers-based architectures\n\n## Getting Started\n\n#### Installation\n1. Create a virtual environment\n\n ```console\n python3 -m venv nlpipesenv\n source nlpipesenv/bin/activate\n ```\n\n2. Install the package\n\n ```console\n pip install nlpipes\n ```\n\n#### Tutorials\n\nHere are some examples on open datasets that show how to use `NLPipes` on different tasks:\n\nName|Notebook|Description|Task|Size|Memory| \n----|-----------|-----|---------|---------|---------|\nGooglePlay Sentiment Detection|[Open](https://git.irt-systemx.fr/ec3-fa16/nlpipes/-/blob/dev/notebooks/googleplay_sentiment_labeling.ipynb)|Train a model to detect the sentiment polarity from the GooglePlay store |Single-label classification|  |  \nStackOverflow tags Detection|Coming soon|Train a model to detect tags from the StackOverFlow questions |Multiple-labels classification|  | \nGooglePlay Aspect and Sentiment Detection|Coming soon|Train a model to detect the aspects from GooglePlay store reviews along their assocated sentiment polarity |Aspect-based classification|  | \n\n\n\n\n#### Why should I use NLPipes?\n- End to End solution for practical text classification problems, especially in the context of aspect based sentiment analysis.\n- Low barrier to entry. Given the unified API for all classification task, there is just one abstraction to learn.\n- Flexible architecture. Most of nlpipes functionnalities are based on callbacks function. This allows new research to be implemented without having to increase the complexity of the core. \n\n\n#### Why shouldn’t I use NLPipes?\n- Nlpipes is not intended to provide building blocks to create customized transformer-based neural network.\n- Speed and mermory footprint efficiency will improve in the future, but if you\'re looking for an high performance library for training and inference, there is probably better alternatives.\n\n\n\n## Notice\n`NLPipes` is still in its early stage and not yet suitable for production usage. The library comes with no warranty as future releases could bring substantial API and behavior changes.\n\n\n<div>Logo created with <a href="https://www.designevo.com/" title="Free Online Logo Maker">DesignEvo logo maker</a></div>\n',
    'author': 'Ayhan UYANIK',
    'author_email': 'ayhan.uyanik@renault.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os
import logging

from typing import List
from typing import Optional
from typing import Union

from pathlib import Path

from dataclasses import dataclass

import tensorflow as tf
import transformers

from transformers import (
    PretrainedConfig,
    PreTrainedTokenizer,
    TFPreTrainedModel,
)

from nlpipes.trainers.trainers import Trainer

from nlpipes.losses.losses import softmax_cross_entropy

from nlpipes.data.data_loaders import DataLoader
from nlpipes.data.data_processors import LanguageModelingDataProcessor

from nlpipes.data.data_selectors import DataSelector
from nlpipes.data.data_augmentors import VocabAugmentor
from nlpipes.data.data_cleaners import clean

from nlpipes.data.data_utils import (
     create_examples,
     split_examples,
)

from nlpipes.data.data_types import (
     Document,
     Corpus,
     Token,
     InputExample,
)

from nlpipes.callbacks.callbacks import (
     Callback,
     TrainingStep,
     History,
     ModelCheckpoint,
     CSVLogger,
     ProgbarLogger,
     EarlyStopping,
     TimedStopping,
)

transformers.logging.set_verbosity_error()

logger = logging.getLogger('__name__')

#uncomment row below for debugging rust tokenizer
#os.environ["RUST_BACKTRACE"]="full" 


@dataclass
class TFPipelineForMaskedLM():
    """ The `pipeline` for the masked language modeling task.
    There are almost no abstractions. Just three standard
    classes are required to use the pipeline: configuration,
    model and tokenizer. 
    
    Args
    ----------
    model(TFPreTrainedModel):
       The model used for the classification task`.
       It usually composed of a bert layer with a
       classification head on top.
    
    tokenizer(PreTrainedTokenizerFast):
       The rust tokenizer used to convert the input
       text into encoded vectors.
    
    config(PretrainedConfig): 
       The model configuration file.
    """
    
    model: TFPreTrainedModel
    tokenizer: PreTrainedTokenizer
    config: PretrainedConfig
    
    def train(
        self,
        texts: Corpus,
        train_frac: float=0.75,
        test_frac: float=0.25,
        num_epochs: int=2,
        batch_size: int=4,
        lr: float=2e-5,
        min_delta: float = 0.01,
        patience: int = 5,
        time_limit: int = 86400,
        rand_seed: int = 69,
        target_corpus_size: Union[int, float]=0.5,
        target_vocab_size: int=31000,
        similarity_metrics: Optional[List[str]]=['euclidean'],
        diversity_metrics: Optional[List[str]]=['entropy'],
        keep_checkpoints: bool = False,
        save_corpus_dir: str = './sub_corpus',
        checkpoints_dir: str = './checkpoints',
        logs_dir: str = './logs',
    ) -> TFPreTrainedModel:
        """ Train the model on a masked modeling task:
        80% [MASK], 10% random words and 10% same word. """
        
        data_selector = DataSelector(
            name='DataSelection',
            tokenizer=self.tokenizer,
            target_corpus_size=target_corpus_size,
            similarity_metrics=similarity_metrics,
            diversity_metrics=diversity_metrics,
        )
        
        vocab_augmentor = VocabAugmentor(
            name='VocabAugmentation',
            tokenizer=self.tokenizer,
            target_vocab_size=target_vocab_size,
            cased=False,
        )

        clean_texts = [clean(text) for text in texts]
        selected_texts = data_selector(clean_texts)
        new_tokens = vocab_augmentor.get_new_tokens(selected_texts)
        
        self.tokenizer.add_tokens(new_tokens)
        self.model.resize_token_embeddings(len(self.tokenizer))
        os.makedirs(save_corpus_dir, exist_ok=True) 
        corpus_path = os.path.join(save_corpus_dir, 'selected_corpus.txt')         
        Path(corpus_path).write_text('\n'.join(selected_texts))
        
        examples = create_examples(
            texts=selected_texts, 
            labels=None,
            config=self.config,
        )
        
        train_examples, test_examples = split_examples(
            examples=examples,
            train_frac=train_frac,
            test_frac=test_frac, 
            shufffle=True, 
        )

        data_processor = LanguageModelingDataProcessor(
            name='DataProcessingForLM',
            tokenizer=self.tokenizer,
            config=self.config,
            mlm_proba=0.15
        )
        
        train_stream = DataLoader(
            name='TrainDataLoading',
            examples=train_examples,
            batch_size=batch_size,
            data_processor=data_processor,
        )
        
        test_stream = DataLoader(
            name='TestDataLoading',
            examples=test_examples,
            batch_size=batch_size,
            data_processor=data_processor,
        )
        
        training_step = TrainingStep(
            name='BatchTraining',
            model=self.model,
            loss_function=softmax_cross_entropy,
            optimizer=tf.optimizers.Adam(lr),
        )
        
        history = History(
            name='History', 
            training_step=training_step,
            loss_metric = tf.metrics.Mean,
            acc_metric = tf.metrics.CategoricalAccuracy,
        )
        
        model_checkpoint = ModelCheckpoint(
            name='ModelCheckpointing',
            model=self.model, 
            history=history,
            checkpoints_dir=checkpoints_dir,
        )
        
        early_stopping = EarlyStopping(
            name='EarlyStopping', 
            history=history,
            direction='minimize',
            min_delta=min_delta,
            patience=patience,
        )
        
        timed_stopping = TimedStopping(
            name='TimedStopping',
            time_limit=time_limit,
        )
        
        csv_logger = CSVLogger(
            name='CSVLogging',
            logs_dir=logs_dir,
        )
        
        progbar_logger = ProgbarLogger(
            name='ProgbarLogging',
            batch_size=batch_size,
            num_epochs=num_epochs,
            history=history,
            num_samples=sum([
                train_stream.num_examples(),
                test_stream.num_examples(),
            ])
        )
        
        callbacks = [
            training_step,
            history,
            model_checkpoint,
            csv_logger,
            progbar_logger,
            early_stopping,
            timed_stopping,
        ]
        
        trainer = Trainer(
            train_stream=train_stream,
            test_stream=test_stream,
            callbacks=callbacks,
            num_epochs=num_epochs,
        )
        
        trainer.train()
        
        if keep_checkpoints == False:
            shutil.rmtree(checkpoints_dir)
        
        return history
    
    def evaluate(self):
        """ Evaluate the model performance on new data
        according to the user-defined evaluation metric """
        raise NotImplementedError
        
    def predict(self):
        """ Get predictions a new unseen data """
        raise NotImplementedError
    
    def save(self, save_dir: str):
        """ Save the model and the tokenizer on disk """
        
        os.makedirs(save_dir, exist_ok=True)        
        self.model.save_pretrained(save_dir)
        self.tokenizer.save_pretrained(save_dir)
        self.config.save_pretrained(save_dir)
        


from typing import Union

import tensorflow as tf


def softmax_cross_entropy(
    labels: tf.Tensor,
    logits: tf.Tensor,
) -> tf.Tensor:
    """ Computes the softmax cross-entropy single label classification """
    softmax = tf.nn.softmax_cross_entropy_with_logits
    return softmax(labels, logits, axis=-1, name='Loss')


def sigmoid_cross_entropy(
    labels: tf.Tensor,
    logits: tf.Tensor,
) -> tf.Tensor:
    """ Computes the sigmoid cross-entropy for multi label classification """
    sigmoid = tf.nn.sigmoid_cross_entropy_with_logits
    return sigmoid(labels, logits, name='Loss')    
    